// -*- C++ -*-
/*!
 * @file DaqComponentBase.h
 * @brief DAQ Component Base class
 * @date
 * @author Yoshiji Yasu, Kazuo Nakayoshi
 *
 * Copyright (C) 2008
 *     Yoshiji Yasu, Kazuo Nakayoshi
 *     Electronics System Group,
 *     KEK, Japan.
 *     All rights reserved.
 *
 */

#ifndef DAQCOMPONENTBASE_H
#define DAQCOMPONENTBASE_H

#include <iostream>

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

#include "DAQServiceSVC_impl.h"
#include "DAQService.hh"
#include "Timer.h"

namespace DAQMW
{
    class DaqComponentBase:
	public RTC::DataFlowComponentBase
    {

    public:
	DaqComponentBase(RTC::Manager* manager)
	    : RTC::DataFlowComponentBase(manager),
	      m_trans_lock(false), m_DAQServicePort("DAQService"),
	      m_comp_name(COMP_NONAME), m_command(CMD_NOP),
	      m_state(LOADED), m_state_prev(LOADED),
	      m_eventByteSize(0), m_loop(0), m_total_event(0), 
	      m_isOnError(false), 
	      m_isTimerAlarm(false),
	      m_debug(false)
	{
	    mytimer = new Timer(STATUS_CYCLE_SEC);
	}
	
	virtual ~DaqComponentBase()
	{
	    delete mytimer;
	}

    protected:
	static const int DAQ_CMD_SIZE   = 10;
	static const int DAQ_STATE_SIZE =  6;
	static const unsigned int STATUS_CYCLE_SEC = 2;
	typedef int (DAQMW::DaqComponentBase::*DAQFunc)();

	DAQFunc m_daq_trans_func[DAQ_CMD_SIZE];
	DAQFunc m_daq_do_func[DAQ_STATE_SIZE];

	bool m_trans_lock;
	DAQServiceSVC_impl m_daq_service0;
	RTC::CorbaPort m_DAQServicePort;

	Timer* mytimer;
	ComponentName m_comp_name;
	DAQCommand m_command;
	DAQCommand m_old_command;
	DAQLifeCycleState m_state;
	DAQLifeCycleState m_state_prev;
	Status m_status;
	unsigned  m_eventByteSize;        
	long long m_loop;
	long long m_total_event;

	std::string m_err_message;

	bool      m_isOnError;
	bool      m_isTimerAlarm;
	bool      m_debug;
      
	int init_command_port()
	{
	    // Set service provider to Ports
	    m_DAQServicePort.registerProvider("daq_svc", "DAQService", m_daq_service0);
  
	    // Set CORBA Service Ports
	    registerPort(m_DAQServicePort);

	    return 0;
	}

	int set_comp_name(ComponentName name)
	{
	    m_comp_name = name;
	    return 0;
	}

	bool check_trans_lock(){
	    return m_trans_lock;
	}

	void set_trans_lock(){
	    m_trans_lock = true;
	}

	void set_trans_unlock(){
	    m_trans_lock = false;
	}

	int transAction(int command) {
	    return (this->*m_daq_trans_func[command])();
	}

	void doAction(int state){
	    (this->*m_daq_do_func[state])();
	}

	virtual int daq_dummy() = 0;
	virtual int daq_configure()   = 0;
	virtual int daq_unconfigure() = 0;
	virtual int daq_start() = 0;
	virtual int daq_run()   = 0;
	virtual int daq_stop()  = 0;
	virtual int daq_pause() = 0;
	virtual int daq_resume(){return 0;}

	int daq_base_dummy()
	{
	    daq_dummy();
	    set_status(COMP_WORKING);
	    sleep(1);
	    return 0;
	}

	int daq_base_configure()
	{
	    m_total_event = 0;
	    set_status(COMP_WORKING);
	    daq_configure();

	    return 0;
	}

	int daq_base_unconfigure()
	{
	    m_total_event = 0;
	    set_status(COMP_WORKING);
	    daq_unconfigure();

	    return 0;
	}

	int daq_base_start()
	{
	    m_total_event = 0;
	    //m_packet_seq_num = 0;
	    m_loop = 0;
	    set_status(COMP_WORKING);
	    daq_start();

	    return 0;
	}

	int daq_base_stop()
	{

	    if (m_isOnError) {
		m_isOnError = false; /// reset error flag
		std::cerr << "*** Error flag was reset\n";
	    }

	    m_err_message = "";
	    set_status(COMP_WORKING);
	    daq_stop();

	    std::cerr << "event num = " << m_status.event_num << std::endl;
	    return 0;
	}

	int daq_base_pause()
	{
	    set_status(COMP_WORKING);
	    daq_pause();

	    return 0;
	}

	int get_command()
	{
	    m_command = m_daq_service0.getCommand();
	    if (m_debug) {
		std::cerr << "Dispatcher: m_command=" << m_command << std::endl;
	    }

	    return 0;
	}

	int set_done()
	{
	    m_daq_service0.setDone();
	    if (m_debug) {
		std::cerr << "Dispatcher::set_done()\n";
	    }

	    return 0;
	}

	virtual int parse_params( ::NVList* list ) = 0;

	int daq_onError(){
	    m_isOnError = true;
	    if (check_trans_lock()) {
		set_trans_unlock();
	    }
	    std::cerr << "### daq_onError(): ERROR Occured\n";
	    std::cerr << m_err_message << std::endl;
	    set_status(COMP_FATAL);
	    sleep(1);

	    return 0;
	}

	int set_status(CompStatus comp_status)
	{

	    m_status.comp_name = m_comp_name;
	    m_status.state = m_state;
	    m_status.event_num = m_total_event;
	    m_status.comp_status = comp_status;

	    m_daq_service0.setStatus(m_status);
	    
	    return 0;
	}

	int fatal_error_report(CompFatalTypes types, int code)
	{
	    std::cerr << "DaqComponentBase::fatal_error_report\n";
	    m_isOnError = true;

	    std::cerr << "### " << mytimer->getDate() << std::endl;

	    set_status(COMP_FATAL);

	    FatalErrorStatus errStatus;
	    errStatus.fatalTypes = types;
	    errStatus.errorCode  = code;
	    m_daq_service0.setFatalStatus(errStatus);

	    return 0;
	}

#ifdef HEADER
	/**
	 *  Setter and Getter methods for Header and Footer in Event Data block.
	 *
	 */
	
	virtual int set_header(unsigned char* header, unsigned daqId, unsigned modNo, 
			       unsigned data_byte_size) 
	{
	    return 0;
	}

	virtual int set_footer(unsigned char* footer, unsigned addr, unsigned sequence_num) 
	{
	    return 0;
	}

	virtual int set_data_to_OutPortBuf(RTC::TimedOctetSeq* out_data, unsigned char* buf, 
					   unsigned addr, unsigned daqId, unsigned modId,
					   unsigned event_num, unsigned seq_num) 
	{
	    return 0;
	}

	virtual int reset_InPort()
	{
	    return 0;
	}


	virtual bool check_header(unsigned char* header, unsigned int received_byte) {
	    return false;
	}

	virtual bool check_footer(unsigned char* footer) {
	    return false;
	}
#endif

	void init_state_table() 
	{
	    m_daq_trans_func[CMD_CONFIGURE]   = &DAQMW::DaqComponentBase::daq_base_configure;
	    m_daq_trans_func[CMD_START]       = &DAQMW::DaqComponentBase::daq_base_start;
	    m_daq_trans_func[CMD_PAUSE]       = &DAQMW::DaqComponentBase::daq_pause;
	    m_daq_trans_func[CMD_RESUME]      = &DAQMW::DaqComponentBase::daq_resume;
	    m_daq_trans_func[CMD_STOP]        = &DAQMW::DaqComponentBase::daq_base_stop;
	    m_daq_trans_func[CMD_UNCONFIGURE] = &DAQMW::DaqComponentBase::daq_base_unconfigure;

	    m_daq_do_func[LOADED]     = &DAQMW::DaqComponentBase::daq_base_dummy;
	    m_daq_do_func[CONFIGURED] = &DAQMW::DaqComponentBase::daq_base_dummy;
	    m_daq_do_func[RUNNING]    = &DAQMW::DaqComponentBase::daq_run;
	    m_daq_do_func[PAUSED]     = &DAQMW::DaqComponentBase::daq_base_dummy;
	}

	int reset_timer() 
	{
	    mytimer->resetTimer();

	    return 0;
	}

	int clockwork_status_report() 
	{
	    if (mytimer->checkTimer()) {
		m_isTimerAlarm = true;
		set_status(COMP_WORKING);
		mytimer->resetTimer();
	    }
	    return 0;
	}

	int daq_do() 
	{
	    int ret = 0;
	    get_command();
	    
	    bool status = true;

	    if (m_command != CMD_NOP) {
		/// new command has come, chage new state
		switch (m_command) {
		case CMD_CONFIGURE:
		    m_state_prev = LOADED;
		    m_state = CONFIGURED;
		    break;
		case CMD_START:
		    m_state_prev = CONFIGURED;
		    m_state = RUNNING;
		    break;
		case CMD_PAUSE:
		    m_state_prev = RUNNING;
		    m_state = PAUSED;
		    set_trans_lock();
		    break;
		case CMD_RESUME:
		    m_state_prev = PAUSED;
		    m_state = RUNNING;
		    break;
		case CMD_STOP:
		    m_state_prev = RUNNING;
		    m_state = CONFIGURED;
		    set_trans_lock();
		    break;
		case CMD_UNCONFIGURE:
		    m_state_prev = RUNNING;
		    m_state = LOADED;
		    m_isOnError = false;
		    break;
		default:
		    status = false;
		    break;
		}

		if (status) {
		    while (check_trans_lock()) {
			if (m_debug) {
			    std::cerr << "### trans locked" << std::endl;
			}
			usleep(0);
			if ( !m_isOnError ) {
			    doAction(m_state_prev);
			}
			else {
			    daq_onError();
			}
		    }
		    ///valid command, do action as transition
		    ret = transAction(m_command);
		    set_status(COMP_WORKING);
		}
		else {
		    std::cerr << "daq_do: transAction call: illegal command" 
			      << std::endl;
		}
		set_done();
	    } else {
		///same command as previous, stay same state, do same action
		if ( !m_isOnError ) {
		    doAction(m_state);
		    clockwork_status_report();
		}
		else {
		    daq_onError();
		}
	    }
	    return ret;
	} /// daq_do()
    }; /// class
} /// namespace

#endif
