// -*- C++ -*-
/*!
 * @file  Gatenet.cpp
 * @brief Gatenet component
 * @date
 * @author Kazuo Nakayoshi <kazuo.nakayoshi@kek.jp>
 *
 * Copyright (C) 2008
 *     Kazuo Nakayoshi
 *     Electronics System Group,
 *     KEK, Japan.
 *     All rights reserved.
 *
 */

#include "Gatenet.h"
#include "stringUtils.h"

int debug = 0; ///debugging flag for pmctl(Gatenet register access)

///const std::string Gatenet::GATENET_ADDR   = "192.168.0.15";
const std::string Gatenet::REGISTER_GATE  = "gate";
const std::string Gatenet::REGISTER_DATE  = "current_time";
const std::string Gatenet::REGISTER_PIDC  = "pulse_id_cnt";
const std::string Gatenet::REGISTER_EXCL  = "extclock";
const std::string Gatenet::VAL_GATE_OPEN  = "0x00800080";
const std::string Gatenet::VAL_GATE_CLOSE = "0x00800000";
const std::string Gatenet::VAL_PIDC_RESET = "0x0000000000";
const std::string Gatenet::CONDITION_FILE = "./condition.json";

// Module specification
static const char* gatenet_spec[] =
{
    "implementation_id", "Gatenet",
    "type_name",         "Gatenet",
    "description",       "Gatenet component",
    "version",           "1.0",
    "vendor",            "Kazuo Nakayoshi, KEK",
    "category",          "example",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    ""
};

Gatenet::Gatenet(RTC::Manager* manager)
    : DAQMW::MlfComponent(manager),
      m_moduleNo(0),
      m_pulIdInitVal(""),
      m_exTrgInitVal(""),
      m_gatenetAddr("192.168.0.15"),
      m_usePulseInitVal(false),
      m_useExTrgInitVal(false),
      m_debug(false)
{
    init_command_port();
    init_state_table( );
    set_comp_name(COMP_GATENET);
}

Gatenet::~Gatenet()
{
}


RTC::ReturnCode_t Gatenet::onInitialize()
{
    if (m_debug) {
	std::cerr << "Gatenet::onInitialize()" << std::endl;
    }

    return RTC::RTC_OK;
}

RTC::ReturnCode_t Gatenet::onExecute(RTC::UniqueId ec_id)
{
    daq_do();

    return RTC::RTC_OK;
}

int Gatenet::daq_dummy() 
{
    return 0;
}

int Gatenet::daq_configure() 
{
    std::cerr << "*** Gatenet::configure" << std::endl;

    ::NVList* paramList;
    paramList = m_daq_service0.getCompParams();
    parse_params(paramList);


    if (adjust_date() != 0) {
        std::cerr << "### ERROR: Gatenet: adjust date failed\n";
        fatal_error_report(GATENET_ADJUST_ERROR, -1);
    }

    return 0;
}

int Gatenet::parse_params(::NVList* nvlist)
{

    int len = (*nvlist).length();
    if (m_debug) {
        std::cerr << "list length:" << (*nvlist).length() << std::endl;
    }

    std::string pidcInitVal;

    for (int i = 0; i < len; i+= 2) {
        std::string sname  = (std::string)(*nvlist)[i].value;
        std::string svalue = (std::string)(*nvlist)[i+1].value;
        if (m_debug) {
            std::cerr << "sname: "   << sname << "  ";
            std::cerr << "value: " << svalue << std::endl;
        }
/*
        if ( sname == "pidcInitVal" ) {
            m_pidc_init_val = svalue;
	    std::cerr << "Gatenet: Initial Pulse ID: " << m_pidc_init_val << std::endl;
        }
*/
	if ( sname == "gatenetAddr") {
	    m_gatenetAddr = svalue;
	    std::cerr << "GATENET ADDR: " << m_gatenetAddr << std::endl;
	}
    }

    return 0;
}


int Gatenet::daq_unconfigure() 
{
    std::cerr << "*** Gatenet::unconfigure" << std::endl;

    return 0;
}

int Gatenet::daq_start()
{
    std::cerr << "*** Gatenet::start" << std::endl;

    if (get_condition() != 0) {
        std::cerr << "### ERROR: Gatenet: get condition failed\n";	
    }

    if (m_usePulseInitVal) {
	if (init_pidc() != 0) {
	    std::cerr << "### ERROR: Gatenet: pulse id init failed\n";
	    fatal_error_report(GATENET_OPEN_ERROR, -1);
	}
    }

    if (m_useExTrgInitVal) {
	if (init_excl() != 0) {
	    std::cerr << "### ERROR: Gatenet: ext. trg init failed\n";
	    fatal_error_report(GATENET_OPEN_ERROR, -1);
	}
    }

    if (open_gate() != 0) {
        std::cerr << "### ERROR: Gatenet: gate open failed\n";
        fatal_error_report(GATENET_OPEN_ERROR, -1);
    }

    return 0;
}

int Gatenet::daq_stop() 
{
    std::cerr << "*** Gatenet::stop" << std::endl;

    if (close_gate() != 0) {
        std::cerr << "### ERROR: Gatenet: gate close failed\n";
        fatal_error_report(GATENET_CLOSE_ERROR, -1);	
    }

    return 0;
}

int Gatenet::daq_pause()
{
    std::cerr << "*** Gatenet::pause" << std::endl;
    if (m_isOnError) {
        m_isOnError = false; /// reset error flag
        m_status.comp_status = COMP_WORKING;

        std::cerr << "*** Error flag was reset\n";
    }

    if (close_gate() != 0) {
        std::cerr << "### ERROR: Gatenet: gate close failed\n";
        fatal_error_report(GATENET_CLOSE_ERROR, -1);	
    }

    return 0;
}

int Gatenet::daq_resume()
{
    if (open_gate() != 0) {
        std::cerr << "### ERROR: Gatenet: gate open failed\n";
        fatal_error_report(GATENET_OPEN_ERROR, -1);
    }

    return 0;
}

int Gatenet::daq_run() 
{
    if (m_debug) {
	std::cerr << "*** Gatenet::run" << std::endl;
    }
    if ( check_trans_lock() ) {
	std::cerr << "Gatherer: STOP command received\n";
	set_trans_unlock();
    }

    sleep(1);
 
   return 0;
}

int Gatenet::get_condition()
{
    int ret = 0;
    Condition mycondition;

    m_usePulseInitVal = false;
    m_useExTrgInitVal = false;

    try {

        Json2ConList json2conlist;
        conList cList;

        if (json2conlist.makeConList(CONDITION_FILE, &cList) == false) {
            std::cerr << "### ERROR: Failed to read the Condition file"
                      << std::endl;
            std::cerr << "### Condition file: " << CONDITION_FILE
                      << std::endl;
        }

	std::string delim = "_";
	std::string myprefix = "common_Gatenet_";

	mycondition.init(&cList);
	mycondition.setPrefix(myprefix);

	int ival;
	std::stringstream initPulsId, initExTrgId;

	if ( mycondition.find("initialPulseId", &ival)) {
	    std::cerr << "find: initialPulseId: " << ival << std::endl;
	    initPulsId << ival;
	    m_pulIdInitVal = initPulsId.str();
	    m_usePulseInitVal = true;
	}

	if ( mycondition.find("initialExTrgId", &ival)) {
	    std::cerr << "find: initialExTrgId: " << ival << std::endl;
	    initExTrgId << ival;
	    m_exTrgInitVal = initExTrgId.str();
	    m_useExTrgInitVal = true;
	}
	if(m_usePulseInitVal) {
	    std::cerr << "initialPulseId:" << m_pulIdInitVal << std::endl;
	}
	if(m_useExTrgInitVal) {
	    std::cerr << "initialExTrgId:" << m_exTrgInitVal << std::endl;
	}

    }
    catch(...) {
	std::cerr << "### Gatherer::set_condition(): exception occured\n";
    }

    return ret;
}

int Gatenet::adjust_date()
{
    int ret = 0;
    int value_count = 0;
    unsigned char recv_buf[MAX_BUF_BYTE];
    unsigned char send_buf[MAX_BUF_BYTE];

    char *value_array[MAX_NUM_VALUE];
    register_info* ri;

    std::cerr << "*** Gatenet: adjust_date()\n";
    if ( (ri = lookup_register((char*)(REGISTER_DATE.c_str()), 
			       strlen((char*)(REGISTER_DATE.c_str())),
			       register_info_list)) == NULL) 
    {
	std::cerr << "### ERROR: Gatenet::adust_date(): lookup_register()\n";
	return -1;
    }

    ri->enc(send_buf, ri->length, value_array, value_count);
    ///if (pm_write_register((char*)(GATENET_ADDR.c_str()), ri, send_buf) != 0) {
    if (pm_write_register((char*)(m_gatenetAddr.c_str()), ri, send_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::adust_date(): pm_wirte_register()\n";
	return -1;
    }

    ///if (pm_read_register( (char*)(GATENET_ADDR.c_str()), ri, recv_buf) != 0) {
    if (pm_read_register( (char*)(m_gatenetAddr.c_str()), ri, recv_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::adust_date(): pm_read_register()\n";
	return -1;
    }

    if (ri->verify(send_buf, recv_buf, ri) != 0) {
	std::cerr << "### ERROR: Gatenet::ajust_date(): verify()\n";
	return -1;
    }

    return ret; 
}

int Gatenet::open_gate()
{ 
    std::cerr << "*** Gatenet: open_gate()\n";

    int ret = 0;
    int value_count = 0;
    unsigned char recv_buf[MAX_BUF_BYTE];
    unsigned char send_buf[MAX_BUF_BYTE];
    char* value_array[MAX_NUM_VALUE];
    register_info* ri;

    if ( (ri = lookup_register((char*)REGISTER_GATE.c_str(), strlen(REGISTER_GATE.c_str()),
			       register_info_list)) == NULL) {
	std::cerr << "### ERROR: Gatenet::open_gate(): lookup_register()\n";
	return -1;
    }

    value_array[value_count] = strdup(VAL_GATE_OPEN.c_str());
    value_count++;
    ri->enc(send_buf, ri->length, value_array, value_count);

    ///if (pm_write_register((char*)(GATENET_ADDR.c_str()), ri, send_buf) != 0) {
    if (pm_write_register((char*)(m_gatenetAddr.c_str()), ri, send_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::open_gate(): pm_wirte_register()\n";
	return -1;
    }

    ///if (pm_read_register( (char*)(GATENET_ADDR.c_str()), ri, recv_buf) != 0) {
    if (pm_read_register( (char*)(m_gatenetAddr.c_str()), ri, recv_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::open_gate(): pm_read_register()\n";
	return -1;
    }

    if (ri->verify(send_buf, recv_buf, ri) != 0) {
	std::cerr << "### ERROR: Gatenet::open_gate(): verify()\n";
	return -1;
    }

    return ret;
}

int Gatenet::close_gate()
{
    std::cerr << "*** Gatenet: close_gate()\n";

    int ret = 0;
    int value_count = 0;
    unsigned char recv_buf[MAX_BUF_BYTE];
    unsigned char send_buf[MAX_BUF_BYTE];
    char* value_array[MAX_NUM_VALUE];
    register_info* ri;

    if ( (ri = lookup_register((char*)REGISTER_GATE.c_str(), strlen(REGISTER_GATE.c_str()),
			       register_info_list)) == NULL) {
	std::cerr << "### ERROR: Gatenet::close_gate(): lookup_register()\n";
	return -1;
    }

    value_array[value_count] = strdup(VAL_GATE_CLOSE.c_str());
    value_count++;
    ri->enc(send_buf, ri->length, value_array, value_count);

    ///if (pm_write_register((char*)(GATENET_ADDR.c_str()), ri, send_buf) != 0) {
    if (pm_write_register((char*)(m_gatenetAddr.c_str()), ri, send_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::close_gate(): pm_wirte_register()\n";
	return -1;
    }

    ///if (pm_read_register( (char*)(GATENET_ADDR.c_str()), ri, recv_buf) != 0) {
    if (pm_read_register( (char*)(m_gatenetAddr.c_str()), ri, recv_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::close_gate(): pm_read_register()\n";
	return -1;
    }

    if (ri->verify(send_buf, recv_buf, ri) != 0) {
	std::cerr << "### ERROR: Gatenet::close_gate(): verify()\n";
	return -1;	
    }

    return ret;
}

int Gatenet::reset_pidc()
{
    std::cerr << "*** Gatenet: reset_pidc()\n";

    int ret = 0;
    int value_count = 0;
    //unsigned char recv_buf[MAX_BUF_BYTE];
    unsigned char send_buf[MAX_BUF_BYTE];
    char* value_array[MAX_NUM_VALUE];
    register_info* ri;

    if ( (ri = lookup_register((char*)REGISTER_PIDC.c_str(), strlen(REGISTER_PIDC.c_str()),
			       register_info_list)) == NULL) {
	std::cerr << "### ERROR: Gatenet::close_gate(): lookup_register()\n";
	return -1;
    }

    value_array[value_count] = strdup(VAL_PIDC_RESET.c_str());
    value_count++;
    ri->enc(send_buf, ri->length, value_array, value_count);

    ///if (pm_write_register((char*)(GATENET_ADDR.c_str()), ri, send_buf) != 0) {
    if (pm_write_register((char*)(m_gatenetAddr.c_str()), ri, send_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::reset_pidc(): pm_wirte_register()\n";
	return -1;
    }

    return ret;
}

int Gatenet::init_pidc()
{
    std::cerr << "*** Gatenet: init_pidc()\n";

    int ret = 0;

    int value_count = 0;
    //unsigned char recv_buf[MAX_BUF_BYTE];
    unsigned char send_buf[MAX_BUF_BYTE];
    char* value_array[MAX_NUM_VALUE];
    register_info* ri;

    if ( (ri = lookup_register((char*)REGISTER_PIDC.c_str(), strlen(REGISTER_PIDC.c_str()),
			       register_info_list)) == NULL) {
	std::cerr << "### ERROR: Gatenet::init_pidc(): lookup_register()\n";
	return -1;
    }


    std::string pidc_init_val = strDec2hex(m_pulIdInitVal);
    std::cerr << "Gatenet::init_pidc: pidc = " 
	      << m_pulIdInitVal
	      << "(" << pidc_init_val << ")" 
	      << std::endl;
    if (pidc_init_val == "") {
	std::cerr << "### ERROR: Gatenet::init_pidc(): invalid initial value\n";
	return -1;
    }
    

    value_array[value_count] = strdup(pidc_init_val.c_str());
    value_count++;
    ri->enc(send_buf, ri->length, value_array, value_count);

    if (pm_write_register((char*)(m_gatenetAddr.c_str()), ri, send_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::init_pidc(): pm_wirte_register()\n";
	return -1;
    }

    return ret;
}

int Gatenet::init_excl()
{
    std::cerr << "*** Gatenet: init_excl()\n";

    int ret = 0;

    int value_count = 0;
    unsigned char send_buf[MAX_BUF_BYTE];
    char* value_array[MAX_NUM_VALUE];
    register_info* ri;

    if ( (ri = lookup_register((char*)REGISTER_EXCL.c_str(), strlen(REGISTER_EXCL.c_str()),
			       register_info_list)) == NULL) {
	std::cerr << "### ERROR: Gatenet::init_excl(): lookup_register()\n";
	return -1;
    }

    std::string excl_init_val = strDec2hex(m_exTrgInitVal);
    std::cerr << "Gatenet::init_excl: excl = " 
	      << m_exTrgInitVal
	      << "(" << excl_init_val << ")" 
	      << std::endl;
    if (excl_init_val == "") {
	std::cerr << "### ERROR: Gatenet::init_excl(): invalid initial value\n";
	return -1;
    }

    value_array[value_count] = strdup(excl_init_val.c_str());
    value_count++;
    ri->enc(send_buf, ri->length, value_array, value_count);

    if (pm_write_register((char*)(m_gatenetAddr.c_str()), ri, send_buf) != 0) {
	std::cerr << "### ERROR: Gatenet::init_excl(): pm_wirte_register()\n";
	return -1;
    }

    return ret;
}

extern "C"
{
 
    void GatenetInit(RTC::Manager* manager)
    {
	RTC::Properties profile(gatenet_spec);
	manager->registerFactory(profile,
				 RTC::Create<Gatenet>,
				 RTC::Delete<Gatenet>);
    }
};


