// -*- C++ -*-
/*!
 * @file Gatherer.cpp
 * @brief Event data gathering component 
 * @date
 * @author Kazuo Nakayoshi <kazuo.nakayoshi@kek.jp>
 *
 * Copyright (C) 2008
 *     Kazuo Nakayoshi
 *     Electronics System Group,
 *     KEK, Japan.
 *     All rights reserved.
 *
 */

#include <sys/time.h>
#include "Gatherer.h"
#include "stringUtils.h"

// Module specification
static const char* gatherer_spec[] =
{
    "implementation_id", "Gatherer",
    "type_name",         "Gatherer",
    "description",       "Event data reading component",
    "version",           "1.0",
    "vendor",            "Kazuo Nakayoshi, KEK",
    "category",          "example",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
};

//const std::string Gatherer::GATENET_ADDR   = "192.168.0.15";
const std::string Gatherer::CONDITION_T0_CNT = "T0";
const std::string Gatherer::CONDITION_FILE   = "./condition.json";

Gatherer::Gatherer(RTC::Manager* manager)
    : DAQMW::MlfComponent(manager),
      m_OutPort("gatherer_out", m_out_data),
      m_out_status(BUF_SUCCESS),
      m_select_tout_counter(0),
      m_loop_tm(0),
      m_t0_sampling_modNo(0), /// T0 cnts are sampled using module No.0
      m_t0_counter(0),
      m_t0_max(0),
      m_requested(false),
      m_condStopMode(false),
      m_finished(false),
      m_debug(false)
{
    // Registration: InPort/OutPort/Service
  
    // Set OutPort buffer
    registerOutPort("gatherer_out", m_OutPort);
    
    init_command_port();
    init_state_table();
    set_comp_name(COMP_GATHERER);

    m_eventByteSize = DAQMW::Neunet::EVENT_BYTE_SIZE;

    if (m_debug) {
	std::cerr << "Gatherer Created\n";
    }
}

Gatherer::~Gatherer()
{
    if (m_debug) {
	std::cerr << "Gatherer Deleted\n";
    }
}


RTC::ReturnCode_t Gatherer::onInitialize()
{
    if (m_debug) {
	std::cerr << "Gatherer onInitialize()\n";
    }

    return RTC::RTC_OK;
}

RTC::ReturnCode_t Gatherer::onExecute(RTC::UniqueId ec_id)
{
    daq_do();

    return RTC::RTC_OK;
}

int Gatherer::daq_dummy()
{
    return 0;
}

int Gatherer::daq_configure()
{
    std::cerr << "*** Gatherer::configure" << std::endl;
    m_dataSrcAddr.clear();

    ::NVList* paramList;
    paramList = m_daq_service0.getCompParams();
    parse_params(paramList);

    return 0;
}

int Gatherer::parse_params(::NVList* nvlist)
{
    unsigned int invalid_daqId = 9999;
    m_daqId = invalid_daqId;

    int len = (*nvlist).length();
    if (m_debug) {
	std::cerr << "list length:" << (*nvlist).length() << std::endl;
    }
    std::string sourceAddr;

    for (int i = 0; i < len; i+= 2) {
	std::string sname  = (std::string)(*nvlist)[i].value;
	std::string svalue = (std::string)(*nvlist)[i+1].value;
	if (m_debug) {
	    std::cerr << "sname: "   << sname << "  ";
	    std::cerr << "value: " << svalue << std::endl;
	}
	if ( sname == "daqId" ) {
	    m_daqId = atoi(svalue.c_str());
	    if (m_debug) {
		std::cerr << "Daq ID:" << m_daqId << std::endl;
	    }
	}
	if ( sname == "srcAddr" ) {
	    if (m_debug) { 
		std::cerr << "source addr: " << svalue << std::endl;
	    }
	    m_dataSrcAddr.push_back(svalue);
	}
    }

    for (int i=0; i< (int)m_dataSrcAddr.size(); i++) {
	std::cerr << m_dataSrcAddr.at(i) << std::endl; 
    }
    
    if (m_daqId == invalid_daqId) {
	std::cerr << "### ERROR: Gatherer: DAQ ID was not specified\n";
	fatal_error_report(GATHERER_NO_DAQID, -1);
    }
     
    std::cerr << "parse params exit\n";
    return 0;
}

int Gatherer::daq_unconfigure()
{
    std::cerr << "*** Gatherer::unconfigure" << std::endl;

    return 0;
}

int Gatherer::set_condition()
{
    ConditionGatherer mycondition;

    try {

        Json2ConList json2conlist;
        conList cList;

        //if (json2conlist.makeConList("../condition.json", &cList) == false) {
        if (json2conlist.makeConList(CONDITION_FILE, &cList) == false) {
            std::cerr << "### ERROR: Failed to read the Condition file"
                      << std::endl;
            std::cerr << "### Condition file: " << CONDITION_FILE
                      << std::endl;
        }

        int modNum = m_dataSrcAddr.size(); /// Num. of NEUNETs (including a GATENET, if you use)

        mycondition.init(m_daqId, modNum, &cList);

	set_params(&mycondition);
    }
    catch(...) {
	std::cerr << "### Gatherer::set_condition(): exception occured\n";
    }

    return 0;
}

int Gatherer::set_params(ConditionGatherer *condGatherer)
{
    int mod_cond_num = (int)(condGatherer->m_neunetsParams.size()); /// Num of Cond. params of NEUNETs
    std::cerr << "mod cond num:" << mod_cond_num << std::endl;
    int mod_num = 0;

    //mod_num = get_neunet_num(&m_dataSrcAddr);
    mod_num = m_dataSrcAddr.size();
    std::cerr << "mod num:" << mod_num << std::endl;

    if(mod_cond_num != mod_num) {
	std::cerr << "### ERROR: Gatherer::set_params: " 
		  << "Num. of Cond. params: " << mod_cond_num
		  << " is not equal to num. of NEUNETs: " << mod_num
		  << std::endl;
	fatal_error_report(USER_ERROR1, -1);
	return -1;
    }

    if (m_debug) {
	for(int i = 0; i < mod_num; i++) {
	    std::cerr << condGatherer->m_neunetsParams[i].modNo  << std::endl;
	    std::cerr << condGatherer->m_neunetsParams[i].lld    << std::endl;
	    std::cerr << condGatherer->m_neunetsParams[i].minTof << std::endl;
	    std::cerr << condGatherer->m_neunetsParams[i].maxTof << std::endl;
	}
    }

    for(int i = 0; i < (int)m_dataSrcAddr.size(); i++) {
	std::string ipAddr = m_dataSrcAddr.at(i);

	//if (ipAddr != GATENET_ADDR ) {
	    int modNo = i;
	    for(int j=0; j< mod_cond_num; j++) {
		if(modNo == condGatherer->m_neunetsParams[j].modNo) {
		    std::cerr << "set_params: " << ipAddr << std::endl;
		    set_lld_tof(ipAddr, &condGatherer->m_neunetsParams[j]);
		} //if(modNo)
	    }//for()
	    //}//if(ipAddr)
    }//for

    return 0;
}

int Gatherer::set_lld_tof(std::string addr, neunetParams* neuParams)
{
    int value_count = 3;
    int bufSize = 128;
    unsigned char recv_buf[bufSize];
    unsigned char send_buf[bufSize];
    char* value_array[value_count];
    register_info* ri;

    char* register_gate = "lld";

    m_condStopMode = false;
    m_t0_max = 0;
    std::cerr << "runStopType  :" << neuParams->runStopType  << std::endl;

    if(neuParams->runStopType == CONDITION_T0_CNT ) {
        m_condStopMode = true;
        m_t0_max       = neuParams->runStopValue;
        std::cerr << "runStopValue :" << neuParams->runStopValue << std::endl;
    }

    std::cerr << "modNo :" << neuParams->modNo  << std::endl;
    std::cerr << "lld   :" << neuParams->lld    << std::endl;
    std::cerr << "minTof:" << neuParams->minTof << std::endl;
    std::cerr << "maxTof:" << neuParams->maxTof << std::endl;

    if ( (ri = lookup_register(register_gate, strlen(register_gate),
                               register_info_list)) == NULL) {
      printf("### ERROR: Gatenet::close_gate(): lookup_register()\n");
        return -1;
    }

    stringstream sslld, ssmintof, ssmaxtof;
    sslld << neuParams->lld;
    ssmintof << neuParams->minTof;
    ssmaxtof << neuParams->maxTof;    

    value_array[0] = strdup(sslld.str().c_str());
    value_array[1] = strdup(ssmintof.str().c_str());
    value_array[2] = strdup(ssmaxtof.str().c_str());


    ri->enc(send_buf, ri->length, value_array, value_count);

    if (pm_write_register( (char*)addr.c_str(), ri, send_buf) != 0) {
      printf("### ERROR: Gatenet::close_gate(): pm_wirte_register()\n");
        return -1;
    }
    if (pm_read_register( (char*)addr.c_str(), ri, recv_buf) != 0) {
      printf("### ERROR: Gatenet::open_gate(): pm_read_register()\n");
        return -1;
    }

    if (ri->verify(send_buf, recv_buf, ri) != 0) {
      printf("### ERROR: Gatenet::open_gate(): verify()\n");
        return -1;
    }

    return 0;
}

/*
int Gatherer::get_neunet_num(stringList* dataSrcAddr)
{
    int neunetNum = 0;
    for(int i = 0; i < (int)dataSrcAddr->size(); i++) {
	std::string ipAddr = dataSrcAddr->at(i);
	if (ipAddr != GATENET_ADDR ) {
	    neunetNum++;
	}
    }
    return neunetNum;
}

int Gatherer::get_gatenet_modNo(stringList* dataSrcAddr)
{
    int gatenet_modNo = -1;
    for(int i = 0; i < (int)dataSrcAddr->size(); i++) {
        std::string ipAddr = dataSrcAddr->at(i);
        if (ipAddr == GATENET_ADDR ) {
            gatenet_modNo = i;
            break;
        }
    }
    return gatenet_modNo;
}

int Gatherer::get_t0_sampling_modNo(stringList* dataSrcAddr)
{
    int gatenet_modNo = get_gatenet_modNo(dataSrcAddr);
    if( gatenet_modNo == 0) {
        m_t0_sampling_modNo = 1;
    }
    else {
        m_t0_sampling_modNo = 0;
    }
    return m_t0_sampling_modNo;
}
*/

int Gatherer::daq_start()
{
    std::cerr << "*** Gatherer::start ***" << std::endl;
    m_select_tout_counter = 0;
    m_t0_counter = 0;
    m_finished = false;

    //std::cerr << "gatenet mod NO:" << get_t0_sampling_modNo(&m_dataSrcAddr) << std::endl;
    set_condition();

    m_selector.clearPsdModuleList();
    DAQMW::psd_module_list* myModuleList;

    try {
	///if ( connect_modules(m_selector, m_dataSrcAddr) != -1) {
	if ( connect_modules() != -1) {
	    std::cerr << "*** Gatherer connected all modules\n";
	    myModuleList = m_selector.getAllList();

	    /// request to all NEUNET modules
	    for (int i = 0; i < (int)m_dataSrcAddr.size(); i++) {
		myModuleList->at(i)->requestEvent(REQUEST_EVENT_NUM);
	    }
	    m_requested = true;
	}
	else {
	    std::cerr << "### ERROR: Gatherer: Could not connect all modules\n";
	    fatal_error_report(GATHERER_CANNOT_CONNECT, -1);
	}
    } catch(...) {
	std::cerr << "### exception occured: Gatherer: daq_start()\n";
	std::cerr << "Gatherer ERROR: Could not connect data source\n";
	fatal_error_report(GATHERER_CANNOT_CONNECT, -1);
    }

    return 0;
}

int Gatherer::daq_stop()
{
    std::cerr << "*** Gatherer::stop" << std::endl;
    /*
    if (m_isOnError) {
	m_isOnError = false; /// reset error flag
	m_status.comp_status = COMP_OK;

	std::cerr << "*** Error flag was reset\n";
    }
    */
    disconnect_modules();

    return 0;
}

int Gatherer::daq_pause()
{
    std::cerr << "*** Gatherer::pause" << std::endl;
    if (m_isOnError) {
	m_isOnError = false; /// reset error flag
	m_status.comp_status = COMP_WORKING;

	std::cerr << "*** Error flag was reset\n";
    }
    disconnect_modules();

    return 0;
}

int Gatherer::daq_resume()
{
    std::cerr << "*** Gatherer::resume" << std::endl;

    m_select_tout_counter = 0;
    m_selector.clearPsdModuleList();

    try {
	///if ( connect_modules(m_selector, m_dataSrcAddr) != -1) {
	if ( connect_modules() != -1) {
	    std::cerr << "*** Gatherer connected all modules\n";
	    DAQMW::psd_module_list* myModuleList = m_selector.getAllList();

	    /// request to all NEUNET modules
	    for (int i = 0; i < (int)m_dataSrcAddr.size(); i++) {
		myModuleList->at(i)->requestEvent(REQUEST_EVENT_NUM);
	    }
	    m_requested = true;
	}
	else {
	    std::cerr << "### ERROR: Gatherer: Could not connect all modules\n";
	    fatal_error_report(GATHERER_CANNOT_CONNECT, -1);
	}
    } catch(...) {
	std::cerr << "### exception occured: Gatherer: daq_start()\n";
	std::cerr << "Gatherer ERROR: Could not connect data source\n";

	fatal_error_report(GATHERER_CANNOT_CONNECT, -1);
    }

    return 0;
}

int Gatherer::check_outPort_status(int mystatus)
{
    int ret = 0;

    if (mystatus == BUF_SUCCESS) {
	m_loop++;
	if (m_debug) {
	    if (m_loop%100 == 0) {
		std::cerr << "Gatherer: m_loop = " << m_loop << std::endl;
		std::cerr << "\033[A\r";
	    }
	}
    }
    else if (mystatus == BUF_TIMEOUT) {
	std::cerr << "Gatherer::Time Out occurred..." << std::endl;
    }
    else if (mystatus == BUF_FATAL) {
	std::cerr << "### Gatherer::Fatal error occurred..." << std::endl;
	///fatal_error_report(GATHERER_OUTPORT_ERROR, -1);
	ret = -1;
    }
    return ret;
}

int Gatherer::set_data_to_OutPortBuf(RTC::TimedOctetSeq* out_data, 
				     unsigned char* buf, unsigned addr,
				     unsigned daqId, unsigned modNo, 
				     unsigned event_num, unsigned seq_num) {

    unsigned char header[8];
    unsigned char footer[8];
    
    set_header(header, daqId, modNo, m_eventByteSize*event_num);
    set_footer(footer, addr, seq_num);

    unsigned int total_event_size = m_eventByteSize * event_num;

    if (m_debug) {
	std::cerr << "seq:" << seq_num << " total_size:" 
		  << (total_event_size + HEADER_BYTE_SIZE + FOOTER_BYTE_SIZE) << std::endl;
    }

    out_data->data.length(total_event_size + HEADER_BYTE_SIZE + FOOTER_BYTE_SIZE);

    for (CORBA::ULong i = 0; i < (CORBA::ULong)HEADER_BYTE_SIZE; ++i) {
	out_data->data[i] = header[i];
    }
    
//    for (CORBA::ULong i = 0; i < (CORBA::ULong)total_event_size; ++i) {
//	out_data->data[i + (CORBA::ULong)EVENT_BUF_OFFSET] = *buf;

    for (CORBA::ULong i = 0; i < (CORBA::ULong)total_event_size; ++i) {
        if (modNo == m_t0_sampling_modNo) { /// if module NO is equals to T0 sampling mod No
            if(i%8 == 0) {                  ///a first byte of event data indicate event type
                if( *buf == NEUNET_T0_MAGIC) {
                    m_t0_counter++;
		    if ( m_condStopMode ) { /// T0 stop mode equals true
			if(m_t0_counter >= m_t0_max) {
			    std::cerr << "*** Reached Maximum T0 Number:" << m_t0_max << std::endl;
			    m_finished = true;
			    set_status(COMP_FINISHED);
			}
                    }
                }
            }
        }
        out_data->data[i + (CORBA::ULong)EVENT_BUF_OFFSET] = *buf;
	buf++;
    }
	    
    for (CORBA::ULong i = 0; i < (CORBA::ULong)FOOTER_BYTE_SIZE; ++i) {
	out_data->data[(CORBA::ULong)HEADER_BYTE_SIZE + (CORBA::ULong)total_event_size + i] =
	    footer[i];
    }
    return 0;
}

int Gatherer::daq_run() 
{
    if (m_debug) {
	std::cerr << "*** Gatherer::daq_run() enter\n";
	std::cerr << "\033[A\r";
    }
    int ret = 0;

    if ( check_trans_lock() && !m_requested ) {
        std::cerr << "Gatherer: STOP command received\n";
        set_trans_unlock();
	return ret;
    }

    if (m_finished) {
	//std::cerr << "*** Gatherer: FINESHED\n";
	set_status(COMP_FINISHED);
        usleep(100000); //100ms
        return ret;
    }

    try {

	if (m_out_status == BUF_TIMEOUT) {
	    std::cerr << "Gatherer: write Timeout" << std::endl;
	    m_out_status = m_OutPort.write(m_out_data);
	    
	    if (m_out_status == BUF_TIMEOUT) {
		return ret;
	    }
	}

	DAQMW::psd_module_list* myReadyList;
	myReadyList = m_selector.getReadyList( SELECT_TIMEOUT ); /// select()

	if (m_debug) {
	    std::cerr << "ready list size:" << m_readyList->size() << std::endl;
	}

	for (int i = 0; i < (int)myReadyList->size(); i++) {

	    unsigned event_num = myReadyList->at(i)->readEventNum();
	    if (m_debug) {
		std::cerr << "event num = " << event_num << std::endl;
	    }
	    if (event_num > 0) {
		if (event_num > DAQMW::Neunet::MAX_EVENT_NUM) { /// TOO MANY DATA ERROR

		    int ipAddr4th = myReadyList->at(i)->getIpAddress()->at(3);

		    std::cerr << "### ERROR: Too many data received:" << event_num << std::endl;
		    std::cerr << "module No.:" << (unsigned)myReadyList->at(i)->getModuleNumber() << std::endl;
		    std::cerr << "IP Address:" << (unsigned)myReadyList->at(i)->getIpAddress()->at(3) 
			      << std::endl;
		    m_err_message = "Gatherer: Too many data received";
		    fatal_error_report(GATHERER_TOO_MANY_DATA, ipAddr4th);
		    
		    return ret;
		}

		unsigned char* buf = myReadyList->at(i)->readEvent(event_num);

		if ( buf != 0 ) { /// valid buffer pointer

		    unsigned int ipAddr4thOct = (unsigned)myReadyList->at(i)->getIpAddress()->at(3);
		    if (m_debug) {		
			std::cerr << "address: " << ipAddr4thOct << std::endl;
		    }
		    unsigned int modNo = (unsigned)myReadyList->at(i)->getModuleNumber();

		    set_data_to_OutPortBuf(&m_out_data, buf, ipAddr4thOct, m_daqId, modNo, 
					   event_num, m_loop);
		    m_out_status = m_OutPort.write(m_out_data); /// send data to next component
		    if (check_outPort_status(m_out_status) == BUF_FATAL) {
			std::cerr << "### Gatherer: OutPort.write(): FATAL ERROR\n";
			fatal_error_report(GATHERER_OUTPORT_ERROR, -1);
			return ret;
		    }

		    m_total_event += event_num;
		}
		else { /// buf == 0, invalid buffer pointer
		    std::cerr << "### Gatherer: socket read(): FATAL ERROR\n";
		    fatal_error_report(GATHERER_SOCK_READ_ERROR, -1);
		    return ret;
		}

	    }//if (event_num > 0)
	    m_requested = false;

	    if (!check_trans_lock()) {
		if (!m_finished) {
		    myReadyList->at(i)->requestEvent(REQUEST_EVENT_NUM); /// event_num = 0;
		    m_requested = true;
		}
	    }
	}//for
    }
    catch(...) {
	fatal_error_report(GATHERER_SOCK_READ_ERROR, -1);

	std::cerr << "### Exception occured: Gatherer: daq_run()\n";
    }

    //if ( check_trans_lock() && !m_requested ) {
    //   std::cerr << "Gatherer: STOP command received\n";
    //   set_trans_unlock();
    //}
   
    //m_loop_tm = (int)(m_stop_tm.tv_sec - m_start_tm.tv_sec) * 1000000
    //    + (int)(m_stop_tm.tv_usec - m_start_tm.tv_usec);

    if (m_debug) {
	std::cerr << "loop time(usec):" << m_loop_tm << std::endl;
    }
    return ret;
}


///int Gatherer::connect_modules(DAQMW::Selector& m_selector, stringList myList)
int Gatherer::connect_modules()
{
    try {
	for (int i=0; i< (int)m_dataSrcAddr.size(); i++) {
	    DAQMW::Neunet *myNeunet = new DAQMW::Neunet();

#ifdef USE_MODNO_IN_CONFIG
	    std::vector< std::string > addModList;
	    split(m_dataSrcAddr.at(i), '/', addModList);
	    /// module NOs in config.xml are used.
	    int modNum = atoi(addModList[1].c_str());

	    myNeunet->init(addModList[0], 
			   NEUNET_BUF_BYTE_SIZE, modNum);
#else       
	    /// module NO automatic numbering. 
	    myNeunet->init(m_dataSrcAddr.at(i), 
			   NEUNET_BUF_BYTE_SIZE, i);
#endif
	    int ret = 0;
	    std::cerr << "connecting: " << m_dataSrcAddr.at(i) << std::endl;
	    ret = myNeunet->connectTcp();

	    if (m_debug) {	
		std::cerr << "connectTCP ret = " << ret << std::endl;
	    }
	    if (ret != 0) {
		if (ret == DAQMW::Sock::ERROR_FATAL) {
		    std::cerr << "### Fatal error: ";
		} else if (ret == DAQMW::Sock::ERROR_TIMEOUT) {
		    std::cerr << "### Timeout: ";
		}
		std::cerr << "could not connect: " << m_dataSrcAddr.at(i) << std::endl;
		return -1;
	    }
	    std::cerr << "connected: " << m_dataSrcAddr.at(i) << std::endl;
	    m_selector.add(myNeunet);
	}
    } catch(...) {
	std::cerr << "### ERROR occured: Gatherer: connect_modules\n";
    }

    return 0;
}

int Gatherer::disconnect_modules()
{
    DAQMW::psd_module_list* myList;
    myList = m_selector.getAllList();

    for (int i=0; i< (int)myList->size(); i++) {
	myList->at(i)->disconnectTcp();
	std::cerr << "disconnected: " 
		  << m_dataSrcAddr.at(i) << std::endl;
    }

    return 0;
}

extern "C"
{
    void GathererInit(RTC::Manager* manager)
    {
	RTC::Properties profile(gatherer_spec);
	manager->registerFactory(profile,
				 RTC::Create<Gatherer>,
				 RTC::Delete<Gatherer>);
    }
};


