// -*- C++ -*-
/*!
 * @file Logger.cpp
 * @brief Event data logging component.
 * @date
 * @author Kazuo Nakayoshi <kazuo.nakayoshi@kek.jp>
 *
 * Copyright (C) 2008
 *     Kazuo Nakayoshi
 *     Electronics System Group,
 *     KEK, Japan.
 *     All rights reserved.
 *
 */

#include "Logger.h"
#include "stringUtils.h"

extern void toLower(std::basic_string<char>& s);

// Module specification
static const char* logger_spec[] =
{
    "implementation_id", "Logger",
    "type_name",         "Logger",
    "description",       "Event data Logging component",
    "version",           "1.0",
    "vendor",            "Kazuo Nakayoshi, KEK",
    "category",          "example",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    ""
};

Logger::Logger(RTC::Manager* manager)
    : DAQMW::MlfComponent(manager),
      m_InPort("logger_in", m_in_data, INPORT_BUF_LEN),
      m_isDataLogging(false),
      m_filesOpened(false),
      m_in_status(BUF_SUCCESS),
      m_update_rate(100),
      m_debug(false)
{
    // Registration: InPort/OutPort/Service
    registerInPort("logger_in", m_InPort);

    init_command_port();
    init_state_table( );
    set_comp_name(COMP_LOGGER);
}

Logger::~Logger()
{
}

RTC::ReturnCode_t Logger::onInitialize()
{
    if (m_debug) {
	std::cerr << "Logger::onInitialize()" << std::endl;
    }

    return RTC::RTC_OK;
}

RTC::ReturnCode_t Logger::onExecute(RTC::UniqueId ec_id)
{
    daq_do();

    return RTC::RTC_OK;
}

int Logger::daq_dummy() 
{
    return 0;
}

int Logger::daq_configure() 
{
    std::cerr << "*** Loggqer::configure" << std::endl;
    int ret = 0;
    m_total_event = 0;
    m_isDataLogging = false; 
    m_moduleNameList.clear();

    ::NVList* list = m_daq_service0.getCompParams();
    parse_params(list);

    return ret;
}

int Logger::parse_params(::NVList* list)
{
    int ret = 0;
    unsigned int invalid_daqId = 9999;
    m_daqId = invalid_daqId;

    bool isExistParamLogging = false;
    bool isExistParamDirName = false;
    bool isExistParamInstId  = false;

    int length = (*list).length();
    for(int i=0; i< length ; i+=2) {
	if (m_debug) {
	    std::cerr << "NVList[" << (*list)[i].name 
		      << ","<< (*list)[i].value << "]" << std::endl;
	}
	std::string sname  = (std::string)(*list)[i].value;
	std::string svalue = (std::string)(*list)[i+1].value;
	if ( sname == "eventByteSize" ) {
	    m_eventByteSize = atoi( svalue.c_str() );
	    std::cerr << "m_eventByteSize:" << m_eventByteSize << std::endl;
	}
	if ( sname == "isLogging" ) {
	    toLower(svalue); // all characters of cvale is converted to lower case.
	    isExistParamLogging = true;
	    if ( svalue == "yes" ) {
		m_isDataLogging = true;
		fileUtils = new FileUtils();
		isExistParamLogging = true;
		std::cerr << "Logger: Save to file: true\n";		
	    }
	    else if (svalue == "no" ) {
		m_isDataLogging = false;
		isExistParamLogging = true;
		std::cerr << "Logger: Save to file: false\n";
	    }
	}

        if ( sname == "monRate" ) {
            m_update_rate = atoi(svalue.c_str());
            if (m_debug) {
                std::cerr << "update rate:" << m_update_rate << std::endl;
            }
        }

    }

    if (m_isDataLogging) {
	
	int modNo = 0;
	for(int i=0; i< length ; i+=2) {
	    if (m_debug) {
		std::cerr << "NVList[" << (*list)[i].name 
			  << ","<< (*list)[i].value << "]" << std::endl;
	    }
	    std::string sname  = (std::string)(*list)[i].value;
	    std::string svalue = (std::string)(*list)[i+1].value;

	    if ( sname == "dirName" ) {
		isExistParamDirName = true;
		m_dirName = svalue;
		if (m_isDataLogging) {
		    std::cerr << "Dir name for data saving:" 
			      << m_dirName << std::endl;
		    ret = fileUtils->check_dir(m_dirName);
		    if (ret != 0) {
			delete fileUtils;
			fileUtils = 0;
			std::cerr << "Can not open directory:" 
				  << m_dirName << std::endl;
			fatal_error_report(LOGGER_BAD_DIR, -1);
			return 0;
		    }
		}
	    }

	    if ( sname == "instId" ) {
		isExistParamInstId = true;
		m_instId = svalue;
		std::cerr << "Instrument ID:" 
			  << m_instId << std::endl;
	    }
	    if ( sname =="runNumber" ) {
		m_runNumber = atoi(svalue.c_str());
		std::cerr << "Run Number:" 
			  << m_runNumber << std::endl;
	    }
	    if ( sname == "daqId" ) {
		m_daqId_str = svalue;
		m_daqId = atoi(svalue.c_str());
		std::cerr << "Daq ID:" 
			  << m_daqId << std::endl;
	    }
	    if ( sname == "maxFileSizeInMegaByte" ) {
		m_maxFileSizeInMByte = strtoul(svalue.c_str(), NULL, 0);
		std::cerr << "Max File size(MByte):" 
			  << m_maxFileSizeInMByte << std::endl;
	    }
	    if ( sname == "srcAddr" ) {
#ifdef USE_MODNO_IN_CONFIG
		std::vector< std::string > addModList;
		split(svalue, '/', addModList);
		m_moduleNameList.push_back(addModList[1]);
#endif
		std::stringstream ss;
		ss << modNo;
		m_moduleNameList.push_back(ss.str());
		modNo++;
	    }
	}
    }

    if (m_isDataLogging && m_daqId == invalid_daqId) {
	std::cerr << "### ERROR: Gatherer: DAQ ID was not specified\n";
	fatal_error_report(LOGGER_NO_DAQID, -1);
    }

    for(int i = 0; i < (int)m_moduleNameList.size(); i++) {
	std::cerr << "modNo=" << m_moduleNameList.at(i) << std::endl;
    }
    return ret;
}

int Logger::daq_unconfigure() 
{
    std::cerr << "*** Logger::unconfigure" << std::endl;
    if (m_isDataLogging) {
	for(int i = 0; i< (int)m_file_list.size(); i++) {
	    delete m_file_list[i];
	}
	m_file_list.clear();
      
	delete fileUtils;
	if (m_debug) {
	    std::cerr << "fileUtils deleted\n";
	}
	fileUtils = 0;
    }
    return 0;
}

int Logger::daq_start() 
{
    std::cerr << "*** Logger::start" << std::endl;

    m_in_status = BUF_SUCCESS;
    m_filesOpened = false;
    m_runNumber = m_daq_service0.getRunNo();
    std::cerr << "m_runNumber:" << m_runNumber << std::endl;

    if (m_isDataLogging) {
        int ret = 0;
	fileUtils->set_run_no(m_runNumber);
	if ( fileUtils->make_dir(m_dirName, m_instId) != 0) {
	    std::cerr << "### ERROR: Logger: make dir failed\n";
	    fatal_error_report(LOGGER_CANNOT_MAKE_DIR, -1);
	    return 0;
	}
	fileUtils->set_daq_id(m_daqId_str);
	std::cerr << "m_maxFileSizeInMByte:" 
		  << m_maxFileSizeInMByte << std::endl;
	fileUtils->set_max_size_in_MegaBytes(m_maxFileSizeInMByte);
	//fileUtils->set_compress_flag(m_compress); /// not implement
	ret = fileUtils->open_files(m_instId, m_moduleNameList);
	if (ret != 0) {
	    std::cerr << "### ERROR: Logger: open file failed\n";
	    fatal_error_report(LOGGER_CANNOT_OPEN_FILE, -1);
	}
	else {
	  m_filesOpened = true;
	}
  }
  return 0;
}

int Logger::daq_stop() 
{
    std::cerr << "*** Logger::stop" << std::endl;

    if (m_isDataLogging && m_filesOpened) {
	std::cerr << "Logger::stop: close files \n";
	fileUtils->close_files(false);
    }

    reset_InPort();
    return 0;
}

int Logger::daq_pause()
{
    std::cerr << "*** Logger::pause" << std::endl;
    return 0;
}

int Logger::daq_resume()
{
    std::cerr << "*** Logger::resume" << std::endl;
    return 0;
}

int Logger::reset_InPort()
{
    TimedOctetSeq dummy_data;

    int ret = BUF_SUCCESS;
    while( ret == BUF_SUCCESS) {
	ret = m_InPort.read(dummy_data);
    }
    std::cerr << "*** Logger::InPort flushed\n";
    return 0;
}

int Logger::daq_run() {

   m_in_status = m_InPort.read(m_in_data);

   /*
    *  When stop command was recieved and InPort.read() returned TIMEOUT, 
    *  Stop command was accepted and the "RUNNING "state is changed to .
    *  Because of Stop command was issued sequentially upstream of data stream.
    *  InPort.read() TIMEOUT means no data exist in upstream.
    */
   if (m_in_status == BUF_TIMEOUT && check_trans_lock() ) {
       set_trans_unlock();
       return 0;
   }
   /*
    *  No data exist in upstream, return and retry.
    */
   else if (m_in_status == BUF_TIMEOUT) {
       return 0;
   }
   /*
    *  Fatal error occcured in Inport.read(), set error status and retrun.
    */
   else if (m_in_status == BUF_FATAL) {
       fatal_error_report(LOGGER_INPORT_ERROR, -1);

       return 0;
   }

   int block_byte_size = m_in_data.data.length();
   int event_byte_size = block_byte_size - HEADER_BYTE_SIZE - FOOTER_BYTE_SIZE;
   if (m_debug) {
       std::cerr << "event_byte_size w/ header, fooger = " << event_byte_size << std::endl;
   }

   if (event_byte_size == 0) {
       return 0;
   } 

   unsigned char header[HEADER_BYTE_SIZE];
   unsigned char footer[FOOTER_BYTE_SIZE];
   for (int i = 0; i < (int)HEADER_BYTE_SIZE; i++) {
       header[i] = m_in_data.data[i];
       if (m_debug) {
	   std::cerr << std::hex <<  (int)m_in_data.data[i] << std::endl;
       }
   }
   int footer_index = block_byte_size - FOOTER_BYTE_SIZE;

   for (int i = 0; i < (int)FOOTER_BYTE_SIZE; i++) {
       footer[i] = m_in_data.data[footer_index + i];
       if (m_debug) {
	   std::cerr << std::hex << (int)m_in_data.data[footer_index + i] << std::endl;
	   std::cerr << std::dec;
       }
   }

   if ( !check_header(&header[0], event_byte_size) ) {
       std::cerr << "### ERROR: header invalid\n";
       fatal_error_report(HEADER_DATA_MISMATCH, -1);

       return 0;
   }

   if ( !check_footer(&footer[0], m_loop) ) {
       std::cerr << "### ERROR: footer invalid\n";
       fatal_error_report(FOOTER_DATA_MISMATCH, -1);

       return 0;
   }

   m_total_event += ( event_byte_size / m_eventByteSize);
	       
   if (m_isDataLogging) {
       try {
	   int modNo = get_modNo(header);
	   fileUtils->write(modNo, (char *)&m_in_data.data[HEADER_BYTE_SIZE], 
			    event_byte_size);
       }
       catch(...) {
	   std::cerr << "### Logger: ERROR occured at data saving\n";
	   fatal_error_report(LOGGER_CANNOT_WRITE_DATA, -1);

	   return 0;
       }
   }

   m_loop++;

   if(m_debug) {
       if (m_loop%m_update_rate == 0) {
	   std::cerr << "Logger: loop = " << m_loop << std::endl;
	   std::cerr << "\033[A\r";
       }
   }

   return 0;
}

extern "C"
{
    void LoggerInit(RTC::Manager* manager)
    {
	RTC::Properties profile(logger_spec);
	manager->registerFactory(profile,
				 RTC::Create<Logger>,
				 RTC::Delete<Logger>);
    }
};


