// -*- C++ -*-
/*!
 * @file MlfComponent.h
 * @brief DAQ Component for MLF Base class
 * @date
 * @author Kazuo Nakayoshi
 *
 * Copyright (C) 2008
 *     Kazuo Nakayoshi
 *     Electronics System Group,
 *     KEK, Japan.
 *     All rights reserved.
 *
 */

#ifndef MLFBASE_H
#define MLFBASE_H

#include <iostream>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>
#include "DaqComponentBase.h"
#include "DAQServiceSVC_impl.h"
#include "DAQService.hh"

namespace DAQMW
{
    class MlfComponent:
	public DAQMW::DaqComponentBase
    {

    public:
	MlfComponent(RTC::Manager* manager)
	    : DAQMW::DaqComponentBase(manager),
	      m_inst(" "),
	      m_daqId(0),
	      m_runNumber(0)
	{
	
	}
	
	virtual ~MlfComponent()
	{

	}

    protected:
	std::string m_inst;
	unsigned m_daqId;
	unsigned m_runNumber;

	static const unsigned int  HEADER_BYTE_SIZE = 8;
	static const unsigned int  FOOTER_BYTE_SIZE = 8;
	static const unsigned char HEADER_MAGIC     = 0xe7;
	static const unsigned char FOOTER_MAGIC     = 0xcc;
	static const unsigned int  EVENT_BUF_OFFSET = HEADER_BYTE_SIZE;
	
	static const int           INPORT_BUF_LEN   = 128;
	static const unsigned int  CPU_DAQ_ID_DIGIT = 2;
	static const unsigned int  MODULE_ID_DIGIT  = 3;


	int eventByteSizeToNum(int event_byte_size) {
	    //std::cerr << "m_eventByteSize:" << m_eventByteSize << std::endl;
	    if (m_eventByteSize != 0) {
		return ( event_byte_size / m_eventByteSize );
	    }
	    else {
		std::cerr << "ERROR: m_eventByteSize = 0\n";
		return -1;
	    }
	}


	/**
	 *  The data structure transferring between DAQ-Components is
	 *  Header data(8bytes) + Event data(8bytes x num_of_events) + Footer data(8bytes).
	 *
	 *  Header data includes magic number(2bytes), DaqId(byte), ModNo(byte) and data byte size(4bytes) 
	 *  except header and footer.
	 *  Footer data includes magic number(2bytes), IP addr of data source and sequence number(4bytes).
	 *
	 *                dat[0]   dat[1]   dat[2]    dat[3]   dat[4]     dat[5]     dat[6]    dat[7]
	 *  Header        0xe7     0xe7     DaqId     ModNo    siz(24:31) siz(16:23) siz(8:15) siz(0:7)
	 *  Event data1
	 *  ...
	 *  Event dataN
	 *  Footer        0xcc     0xcc     add(8:15) add(0:7) seq(24:31) seq(16:23) seq(8:15) seq(0:7)
	 */

	int set_header(unsigned char* header, unsigned daqId, unsigned ModNo, unsigned data_byte_size) {
	    header[0] = HEADER_MAGIC;
	    header[1] = HEADER_MAGIC;
	    header[2] = daqId;
	    header[3] = ModNo;
	    header[4] = (data_byte_size & 0xff000000) >> 24;
	    header[5] = (data_byte_size & 0x00ff0000) >> 16;
	    header[6] = (data_byte_size & 0x0000ff00) >>  8;
	    header[7] = (data_byte_size & 0x000000ff);
	    return 0;
	}

	int set_footer(unsigned char* footer, unsigned addr, unsigned sequence_num) 
	{
	    if(m_debug) {
		std::cerr << "*** set_footer:SEQ NO: " << sequence_num << std::endl;
	    }
	    footer[0] = FOOTER_MAGIC;
	    footer[1] = FOOTER_MAGIC;
	    footer[2] = (addr & 0xff00) >> 8;
	    footer[3] = (addr & 0x00ff);
	    footer[4] = (sequence_num & 0xff000000) >> 24;
	    footer[5] = (sequence_num & 0x00ff0000) >> 16;
	    footer[6] = (sequence_num & 0x0000ff00) >>  8;
	    footer[7] = (sequence_num & 0x000000ff);
	    return 0;
	}

	int get_modNo(unsigned char* header) {
	    return (int)(header[3]);
	}

	int get_daqId(unsigned char* header) {
	    return (int)(header[2]);
	}

	bool check_header(unsigned char* header, unsigned received_byte) 
	{
	    bool ret = false;

	    if (header[0] == HEADER_MAGIC && header[1] == HEADER_MAGIC) {
		unsigned int event_size = (  header[4] << 24 ) 
   		                         + ( header[5] << 16 )
		                         + ( header[6] <<  8 )
		                         +   header[7];
		if (m_debug) {
		    std::cerr << "event_size:" << event_size     << std::endl;
		    std::cerr << "Module No: " << (int)header[3] << std::endl;
		    std::cerr << "DAQ CPUID: " << (int)header[2] << std::endl;
		}
		if (received_byte == event_size) {
		    ret = true;
		}
		else {
		    std::cerr << "### ERROR: Event byte size missmatch" << std::endl;
		}
	    }
	    else {
		std::cerr << "### ERROR: Bad Magic Num:" << std::hex << header[0] << header[1] << std::endl;
	    }
	    return ret;
	}

	bool check_footer(unsigned char* footer, unsigned loop_cnt) 
	{
	    bool ret = false;
	    if (footer[0] == FOOTER_MAGIC && footer[1] == FOOTER_MAGIC) {
		unsigned int seq_num = (  footer[4] << 24 )
		                      + ( footer[5] << 16 )
		                      + ( footer[6] <<  8 )
		                      +   footer[7];
		if (m_debug) {
		    std::cerr << "seq: " << seq_num << std::endl;
		}
		if (seq_num == loop_cnt) {
		    ret = true;
		}
		else {
		    std::cerr << "### ERROR: Sequence No. missmatch" << std::endl;
		    std::cerr << "sequece no. in footer :" << seq_num << std::endl;
		    std::cerr << "loop cnts at component:" << loop_cnt << std::endl;
		}
	    }
	    return ret;
	}

    }; /// class
} /// namespace

#endif
