// -*- C++ -*-
/*!
 * @file Monitor.cpp
 * @brief Event data monitoring component.
 * @date
 * @author Hiroshi Sendai
 *
 * Copyright (C) 2008 - 2009
 *     Kazuo Nakayoshi
 *     Hiroshi Sendai
 *     KEK, Tsukuba, Ibaraki, Japan
 *     All rights reserved.
 *
 */

#include "Monitor.h"
#include "stringUtils.h"

static const char* monitor_spec[] =
{
    "implementation_id", "Monitor",
    "type_name",         "Monitor",
    "description",       "Data Monitoring component",
    "version",           "1.0",
    "vendor",            "Kazuo Nakayoshi, Hiroshi Sendai, KEK",
    "category",          "example",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
};

Monitor::Monitor(RTC::Manager* manager)
    : DAQMW::MlfComponent(manager),
      m_InPort("monitor_in", m_in_data, INPORT_BUF_LEN),
      m_in_status(BUF_SUCCESS),
	  m_sampling_rate(100),
	  m_update_rate(1000), 
      m_debug(false),
	  m_monitor_debug(false),
	  m_monitor_debug_verbose(false),
	  m_num_of_psd_per_module(8),
	  m_total_module_num(0),
	  m_daqId(0)
{
    if (m_debug) {
        std::cerr << "Monitor constructor: enter" << std::endl;
    }
    registerInPort("monitor_in", m_InPort);

    init_command_port();
    init_state_table();
    set_comp_name(COMP_MONITOR);

	// Prepare event byte size change in the future //
    m_eventByteSize = EVENT_BYTE_SIZE;

    if (m_debug) {    
        std::cerr << "Monitor constructor: exit" << std::endl;
    }
}

Monitor::~Monitor() 
{
}

RTC::ReturnCode_t Monitor::onInitialize()
{
    if (m_debug) {
        std::cerr << "*** Monitor::onInitialize()" << std::endl;
    }
    return RTC::RTC_OK;
}

RTC::ReturnCode_t Monitor::onExecute(RTC::UniqueId ec_id)
{
    daq_do();

    return RTC::RTC_OK;
}

int Monitor::daq_dummy()
{
    return 0;
}

int Monitor::print_monitor_param(void)
{
	//////////////////// Manyo Monitor ////////////////////
	std::cerr << "m_monitorParam.tof_min: "              << m_monitorParam.tof_min
		<< std::endl;
	std::cerr << "m_monitorParam.tof_max: "              << m_monitorParam.tof_max
		<< std::endl;
	std::cerr << "m_monitorParam.num_of_tof_bin: "       << m_monitorParam.num_of_tof_bin
		<< std::endl;
	std::cerr << "m_monitorParam.num_of_pixel_per_psd: " << m_monitorParam.num_of_pixel_per_psd
		<< std::endl;
	std::cerr << "m_monitorParam.monitor_update_rate: "  << m_monitorParam.monitor_update_rate
		<< std::endl;
	std::cerr << "m_num_of_psd_per_module: " << m_num_of_psd_per_module
		<< std::endl;
	std::cerr << "m_total_module_num: "      << m_total_module_num
		<< std::endl;
	std::cerr << "m_gnuplot_path: "          << m_gnuplot_path
		<< std::endl;
	std::cerr << "m_png_output_dir: "        << m_png_output_dir
		<< std::endl;
	std::cerr << "m_monitor_debug: "         << m_monitor_debug
		<< std::endl;
	std::cerr << "m_monitor_debug_verbose: " << m_monitor_debug_verbose
		<< std::endl;
	////////////////// Manyo Monitor End //////////////////

	return 0;
}
	
int Monitor::set_condition()
{
	ConditionMonitor conditionMonitor;
	try {
		conditionMonitor.initialize("./condition.json");
		if (conditionMonitor.getParam("common_Monitor_", &m_monitorParam)) {
			std::cerr << "condition OK" << std::endl;
		}
		else {
			throw "condition error";
		}
	}
	catch (...) {
		std::cerr << "condition error" << std::endl;
	}
	return 0;
}

int Monitor::daq_configure()
{
    std::cerr << "*** Monitor::configure in" << std::endl;

    ::NVList* list = m_daq_service0.getCompParams();
    parse_params(list);

    if (m_debug) {
        std::cerr << "Monitor::configure out" << std::endl;
    }

	////////// Manyo Monitor (Condition) //////////
	set_condition();
	//////// Manyo Monitor (Condition) End ////////

	//////////////////// Manyo Monitor ////////////////////
	print_monitor_param();
	m_gplot = new GnuplotInterface(m_gnuplot_path);
	m_stool = new StringTools();
	m_evs   = new EventDataToHistogramEmulator(
				"abc.mask",
				(Double) m_monitorParam.tof_min,
				(Double) m_monitorParam.tof_max,
				m_monitorParam.num_of_tof_bin,
				m_num_of_psd_per_module,
				m_monitorParam.num_of_pixel_per_psd,
				m_total_module_num
			  );
	m_OnLineGp = new OnLineGnuPlot(m_gnuplot_path);
	m_vg = m_evs->PutGslHistogramVec();
	m_OnLineGp->SetDataObj(
		m_vg,
		m_total_module_num * m_num_of_psd_per_module,
		m_monitorParam.num_of_pixel_per_psd
	);
	////////////////// Manyo Monitor End //////////////////

    return 0;
}

int Monitor::parse_params(::NVList* nvlist)
{
    int len = (*nvlist).length();
    if (m_debug) {
        std::cerr << "list length:" << (*nvlist).length() << std::endl;
    }
    std::string sampRateAndPort;

	//////////////////// Manyo Monitor /////////////////////
	///// reading from config.xml (not condition.json) /////
	m_total_module_num = 0;
    for (int i = 0; i < len; i += 2) {
        std::string sname  = (std::string)(*nvlist)[i].value;
        std::string svalue = (std::string)(*nvlist)[i+1].value;
        if (m_debug) {
            std::cerr << "sname: " << sname  << " ";
            std::cerr << "value: " << svalue << std::endl;
        }
        if (sname == "monRate") {
            m_update_rate = atoi(svalue.c_str());
            if (m_debug) {
                std::cerr << "m_update_rate: " << m_update_rate << std::endl;
            }
        }
        if (sname == "samplingRate") {
            sampRateAndPort = svalue;
            if (m_debug) {
                std::cerr << "Sampling rate of Monitor:" << svalue
                          << sampRateAndPort << std::endl;
            }
        }
		if (sname == "monitor_debug") {
			if (svalue == "YES") {
				m_monitor_debug = true;
			}
		}
		if (sname == "monitor_debug_verbose") {
			if (svalue == "YES") {
				m_monitor_debug_verbose = true;
			}
		}
		if (sname == "gnuplot_path") {
			m_gnuplot_path = svalue;
		}
		if (sname == "png_output_dir") {
			m_png_output_dir = svalue;
		}
		if (sname == "num_of_psd_per_module") {
			m_num_of_psd_per_module = atoi(svalue.c_str());
		}
		if (sname == "daqId") {
			m_daqId = atoi(svalue.c_str());
		}

		// Count the number of NEUNET module
		if (sname == "srcAddr") {
			m_total_module_num ++;
		}
    }
	////////////////// Manyo Monitor End //////////////////

    if (sampRateAndPort != "") {
        std::cerr << "sampRateAndPort" << sampRateAndPort << std::endl;
        std::vector< std::string > ratePortList;
        split(sampRateAndPort, '/', ratePortList);
        m_sampling_rate      = atoi(ratePortList[1].c_str());

	if (m_sampling_rate <= 0) {
	  std::cerr << "ERROR: Monitor: Invalid sampling rate" << m_sampling_rate;
	  m_sampling_rate = 1;
	}
    }
    else {
		// full sampling. All data are sent to Logger and Monitor.
        m_sampling_rate = 1;
    }

    std::cerr << "m_sampling_rate:" << m_sampling_rate      << std::endl;

    return 0;
}

int Monitor::daq_unconfigure() {
    std::cerr<< "*** Monitor::unconfigure" << std::endl;

	//////////////////// Manyo Monitor ////////////////////
	delete m_gplot;
	delete m_stool;
	delete m_OnLineGp;
	delete m_evs;
	////////////////// Manyo Monitor End //////////////////

    return 0;
}

int Monitor::daq_start()
{
    std::cerr<< "*** Monitor::start" << std::endl;

    m_in_status = BUF_SUCCESS;

    return 0;
}

int Monitor::daq_stop() {
    std::cerr << "*** Monitor::stop" << std::endl;

    if (m_debug) {
        std::cerr << "Total Event Counter = " << m_total_event << std::endl;
    }

	//////////////////// Manyo Monitor ////////////////////
    m_evs->Reset();
	////////////////// Manyo Monitor End //////////////////

    reset_InPort();
    return 0;
}

int Monitor::daq_pause()
{
    std::cerr << "*** Monitor::pause" << std::endl;
    return 0;
}

int Monitor::daq_resume()
{
    std::cerr << "*** Monitor::resume" << std::endl;
    return 0;
}

int Monitor::reset_InPort()
{
    TimedOctetSeq dummy_data;

    int ret = BUF_SUCCESS;
    while (ret == BUF_SUCCESS) {
	ret = m_InPort.read(dummy_data);
    }
    std::cerr << "*** Monitor::InPort flushed\n";
    return 0;
}

int Monitor::check_gatherer_header_footer(int comp_seq_num, unsigned char *header, unsigned char *footer)
{

	int block_byte_size = m_in_data.data.length();
	int event_byte_size = block_byte_size - HEADER_BYTE_SIZE - FOOTER_BYTE_SIZE;

    //unsigned char header[HEADER_BYTE_SIZE];
    //unsigned char footer[FOOTER_BYTE_SIZE];

    for (unsigned int i = 0; i < HEADER_BYTE_SIZE; i++) {
        header[i] = m_in_data.data[i];
        if (m_debug) {
            std::cerr << std::hex << (int) m_in_data.data[i] << std::endl;
        }
    }

    int footer_index = block_byte_size - FOOTER_BYTE_SIZE;
    for (unsigned int i = 0; i < FOOTER_BYTE_SIZE; i++) {
        footer[i] = m_in_data.data[footer_index + i];
        if (m_debug) {
            std::cerr << std::hex << (int)m_in_data.data[footer_index + i]
					  << std::endl;
            std::cerr << std::dec;
        }
    }

    //if (check_header(&header[0], event_byte_size) == false) {
    if (check_header(header, event_byte_size) == false) {
		throw HEADER_DATA_MISMATCH;
    }
    //if (check_footer(&footer[0], comp_seq_num) == false) {
    if (check_footer(footer, comp_seq_num) == false) {
		throw FOOTER_DATA_MISMATCH;
    }
	return 0;
}

int Monitor::daq_run()
{
    m_in_status = m_InPort.read(m_in_data);
   /*
    *  When stop command was recieved and InPort.read() returned TIMEOUT, 
    *  Stop command was accepted and the "RUNNING "state is changed to .
    *  Because of Stop command was issued sequentially upstream of data stream.
    *  InPort.read() TIMEOUT means no data exist in upstream.
    */
    if ((m_in_status == BUF_TIMEOUT) && check_trans_lock()) {
		set_trans_unlock();
		return 0;
    }
    /*
     *  No data exist in upstream, return and retry.
     */
    else if (m_in_status == BUF_TIMEOUT) {
                return 0;
    }
    /*
     *  Fatal error occcured in Inport.read(), set error status and retrun.
     */
    else if (m_in_status == BUF_FATAL) {
		fatal_error_report(MONITOR_INPORT_ERROR, -1);
		return 0;
    }
    
    int block_byte_size = m_in_data.data.length();
    int event_byte_size = block_byte_size - HEADER_BYTE_SIZE - FOOTER_BYTE_SIZE;
    if (m_debug) {
        std::cerr << "event_byte_size w/o header, footer = " << event_byte_size 
		  << std::endl;
    }

    if (event_byte_size == 0) {
		return 0;
	}

	////////// Check header and footer appended by Gatherer /////////
    unsigned char header[HEADER_BYTE_SIZE];
    unsigned char footer[FOOTER_BYTE_SIZE];
	try {
	     check_gatherer_header_footer((int)m_loop*m_sampling_rate, header, footer);
	}
	catch (CompFatalTypes compFatalTypes) {
	    fatal_error_report(compFatalTypes, -1);
	    return 0;
	}
	catch (...) {
	    std::cerr << "Unknown Error" << std::endl;
	    return 0;
	}

	// Get modNo, daqId from header that was prepended by GathererComp
	UInt4 modNo = get_modNo(header); // We cannot check modNo.
	UInt4 daqId = get_daqId(header); // Try to check daqId.
	if (m_monitor_debug_verbose) {
		std::cerr << "modNo: " << modNo << std::endl;
		std::cerr << "daqId: " << daqId << std::endl;
	}
	if (m_daqId != daqId) {
		std::cerr << "### ERROR: Monitor: DaqId mismatch.  "
				  << "From config.xml: " << m_daqId << " "
				  << "From header: "     << daqId   << std::endl;
	    fatal_error_report(MONITOR_NO_DAQID, -1);
		return 0;
	}
	////////// Check header and footer appended by Gatherer Done /////////

	////////// Extract event data from m_in_data //////////
    for (int i = 0; i < event_byte_size; i++) {
        m_cdata[i] = m_in_data.data[i + HEADER_BYTE_SIZE];
    }

    m_total_event += eventByteSizeToNum(event_byte_size);
    m_loop++;

	if (m_monitor_debug) {
		if ( (m_loop % m_monitorParam.monitor_update_rate == 0) || m_loop == 100) {
	    	std::cerr << "Monitor: m_loop = " << m_loop << std::endl;
	    	//std::cerr << "\033[A\r";
		}
    }

	//////////////////// Manyo Monitor ////////////////////
	// Increment Event Data to Histogram //
    m_evs->Increment(0, &m_cdata[0], event_byte_size / EVENT_BYTE_SIZE, modNo);
	// Create Histogram PNG Files //
    if (m_loop % m_monitorParam.monitor_update_rate == 0) {
		create_png_files();
    }
	////////////////// Manyo Monitor End //////////////////

    return 0;
}

extern "C" {
    void MonitorInit(RTC::Manager* manager)
    {
        RTC::Properties profile(monitor_spec);
        manager->registerFactory(profile, 
                                 RTC::Create<Monitor>, 
                                 RTC::Delete<Monitor>);
    }
};
