// -*- C++ -*-
#ifndef SELECTOR_H
#define SELECTOR_H
#include <iostream>
#include <vector>
#include <sys/select.h>
#include "PsdModule.h"
#include "Neunet.h"

namespace DAQMW {

    typedef std::vector<int> ready_fd_list;
    typedef std::vector< Neunet* > psd_module_list;

    class PsdModule;
    class Neunet;

    /// exception class for selector function.
    class SelectorException : public std::exception
    {
    public:
	explicit SelectorException(const std::string& msg)
	    : m_msg(msg)
	{
	    m_reason = 0;
	}
	virtual ~SelectorException() throw()
	{
	}
	virtual const char* what() const throw()
	{
	    return m_msg.c_str();
	}
	int reason() const throw()
	{
	    return m_reason;
	}
	
    private:
	int m_reason;
	std::string m_msg;
    };
 
    class Selector
    {
    public:
	Selector()
	    : m_maxfd(0), m_debug(false)
	{
	    clearList();
	}

	virtual ~Selector(){}
	
	int add(Neunet* module)
	{
	    m_allModules.push_back(module);
	    int fd = module->getSockFd();
	    FD_SET(fd, &m_allset);

	    m_maxfd = max(fd, m_maxfd);
	    if (m_debug) {
		std::cerr << "fd = " << fd << std::endl;
		std::cerr << "max fd = " << m_maxfd << std::endl;
	    }
	    return 0;
	}

	int clearList()
	{
	    m_maxfd = 0;
	    FD_ZERO(&m_allset);
	    FD_ZERO(&m_rdset);	    
	    return 0;
	}

	int initReadyList()
	{
	    m_rdset = m_allset;
	    return 0;
	}

	int clearReadyList()
	{
	    m_readyModules.clear();
	    return 0;
	}

	int showAllFds()
	{
	    std::cerr << "**** dump all fds in m_allModules ****\n";
	    for(int i=0; i< (int)m_allModules.size(); i++) {
		std::cerr << "fd =" << (Neunet*)m_allModules[i]->getSockFd() 
			  << std::endl; 
	    }
 
	    return 0;
	}

	int addReadyList(Neunet* module)
	{
	    m_readyModules.push_back( module );
	    return 0;
	}

	psd_module_list* getReadyList(int timeout_sec)
	{
	    initReadyList();
	    initTimeOut(timeout_sec);
	    clearReadyList();

	    int ret = select(m_maxfd+1, &m_rdset, NULL, NULL, &m_timeout);

	    if ( ret < 0) {      /// select() error occured.
		std::cerr << "### ERROR: select ret= " << ret << std::endl;
		perror("ERROR: Select:");
		showAllFds(); 
		throw SelectorException("Selector Fatal error");
	    }
	    else if (ret == 0) {  /// select() timeout occured. 
		std::cerr << "### WORNING: select timeout occured." << std::endl;
		throw SelectorException("### Selector timeout");
	    }

	    for (int i=0; i< (int)m_allModules.size(); i++) {
		int sockfd = m_allModules[i]->getSockFd();
		if ( FD_ISSET(sockfd, &m_rdset)) {
		    addReadyList(m_allModules[i]);
		}
	    }

	    /**
	     * Because of fair chance for the order of reading, the waiting queue list 
	     * should be random shuffled.
	     */
	    std::random_shuffle(m_readyModules.begin(), m_readyModules.end());

	    return &m_readyModules;
	}

	psd_module_list* getAllList()
	{
	    return &m_allModules;
	} 

	int clearPsdModuleList() 
	{
	    m_maxfd = 0;
	    for (psd_module_list::iterator it = m_allModules.begin();
		it != m_allModules.end(); ++it) {
		delete *it;
	    }
	    m_allModules.clear();
	    return 0;
	}

    private:
	int max(int x, int y)
	{
	    return ((x >= y)? x : y );
	}

	int initTimeOut(int timeout_sec) 
	{
	    m_timeout.tv_usec = 0;
	    m_timeout.tv_sec  = timeout_sec;
	    return 0;
	}

	psd_module_list m_allModules;
	psd_module_list m_readyModules;
	int   m_nfds;
	int   m_maxfd;
	fd_set m_allset;
	fd_set m_rdset;
	fd_set m_errorfds;
	struct timeval m_timeout;
	bool m_debug;
    };
}

#endif
